/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


//  xxxxxxxxxxxxxxx     wm14 global variable declarations  xxxxxxxxxxxxxxxx


// These variables below can be directly modified by the user on the options form: 

var item_cnt= 1;       // The number of stimuli in the task

var stype = '1';      //  stype => Stimulus type: "0" => Alphanumeric characters; "1"  => Images (selected by the user)

var alfa_type = '0';  // Type of Alphanumeric characters => "0" = Digits; "1" = Letters-lowercase; "2" = Letters-UPPERCASE; "3" = All digits and letters

var alfa_cnt = 1;  // The number of alphanumeric characters in the task

var font_family = 'Georgia, serif';  // default canvas ctx.fontFamily (see drawText function)

var back_clr= '#ffffff';  // background (canvas) color. Enabled for image stimuli only

var s_parts = 5;       // Raster start size for each stimulus 
  
var demo = false;     // if (demo == true) => Show correct solution (not for Stimulus type = images)

var feedback = true;    // additional information will be displayed at the bottom of the page (true) or not (false)

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global variables:

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_task = false;  // task in progress or not => for html: body onresize="pageResize()"

var item = 0;   // current item in task     

var c_parts;  		// current number of stimulus parts. Range: 1.. s_parts

var stim = [];   // stimulus displayed to the user (alphanumeric characters)

var ans = [];  	//  user (client) answer for each item 

var dv = [];      // for each stimulus (array): the number of parts (raster) on which the user recognized the stimulus

var font_size= 24;   // global fontSize (px) for drawText() function

var bounds={};      // image stimulus bounds on canvas 

var rects = [];     // an array of objects that define different rectangles

var canvas;         // document.getElementById('canvas')

var ctx;						// canvas.getContext('2d');


//  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx


// app init, init page layout, [center panels], and show start panel.. 
function init() 
{      
  if (isStorage() && storageAvailable('localStorage'))  
  {
    document.getElementById("store_group").disabled= false;
  }
  else 
  {
    document.getElementById('store_msg').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline).";
    document.getElementById('store_group').disabled= true;  
    document.getElementById('clr_store').disabled= true;     // IE11 bug  => childrens are not auto disabled
    document.getElementById('store').disabled= true;          // IE11 bug => childrens are not auto disabled
  }    
	
	if ( !readerExists() ) {
	  document.getElementById('stype').disabled= true;
		document.getElementById('no_reader').style.display= 'block';		
	}	
	
	setGlobals();
	  
	setIns();
	
	loadFont();
	
	showStart(); 
 
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 and Edge offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// checks if the client's web browser supports this: FileReader object, Blob object, FileReader.readAsDataURL() 
function readerExists() 
{
	try
	{		
		// create Blob from a typed array and strings: https://javascript.info/blob
		let hello = new Uint8Array([72, 101, 108, 108, 111]); // "Hello" in binary form
		let blob = new Blob([hello, ' ', 'world'], {type: 'text/plain'});		
		let reader = new FileReader();
		if (blob) {
			reader.readAsDataURL(blob);
			return true;
		}
	}	 
  catch(e) { 
    return false;
  }
	return false;
}


// set app global objects (variables) and attach event handlers 
function setGlobals()  {
	canvas= document.getElementById('canvas');
  ctx= canvas.getContext('2d'); 
  document.getElementById('answer').onkeyup = answerKeyUp;	  
}	


// raised when page is resized. see html: <body onresize="pageResize()"
function pageResize()  
{
	if (is_task) { 
		setPage();
	}
}	


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement("option");
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex=1; // Georgia, serif
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[1];
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('child');      
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  if ( document.getElementById('store_group').disabled === false && localStorage.getItem('wm14-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('wm14-v1.0')).length;                                                        
    document.getElementById("store_cnt").textContent=  'You have a total of ' + cnt  + ' wm14 results stored in your web browser.';
  }
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add("child");   
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// set instructions text
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';	
	if (stype === '1')   {
	  document.getElementById('ins_1').style.display= 'block';
		document.getElementById('ins_0').style.display= 'none';
		document.getElementById('ins_1_header').style.display= document.getElementById('ins_header').checked ? 'block' : 'none';
	}
  else  {
		document.getElementById('ins_0').style.display= 'block';
		document.getElementById('ins_1').style.display= 'none';
	}			
}	

  
// refresh Options (update *_msg label) => give user short option HELP
function refOption(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';    
           
	switch (my_id)
	{
		
		case 'item_cnt':
			itemsValid();
			break;
		
		case 'stype':
		  if (e.value === '1')  {
				msg.textContent= 'Images will be displayed (you must select image files)';
				document.getElementById('alfa_opt').style.display= 'none';        // document.getElementById("alfa_opt").disabled= true;
				document.getElementById('images_opt').style.display= 'block';   // 	document.getElementById("images_opt").disabled= false;
				document.getElementById('demo').disabled= true;
				document.getElementById('is_border').disabled= document.getElementById('rotation').checked ? true : false; // enabled only for image with no rotation
			}
			else  {
				msg.textContent= 'Random alphanumeric characters will be displayed'; 
			  document.getElementById('images_opt').style.display= 'none';  // 	document.getElementById("images_opt").disabled= true;
				document.getElementById('alfa_opt').style.display= 'block';    // document.getElementById("alfa_opt").disabled= false;
				document.getElementById('demo').disabled= false;
				document.getElementById('is_border').disabled= true;
			}	
      itemsValid();     		
			break;
		
		case 'alfa_cnt':
			if (e.validity.valid)  {           
				msg.removeAttribute("class");       
				msg.textContent= 'The number of alphanumeric characters in the task';       //msg.textContent=  e.value; 
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			break;
									
		case 's_parts':  
			if (e.validity.valid)  {           
				msg.removeAttribute("class");  
        let tmp= parseInt(e.value, 10) * parseInt(e.value, 10);        
				msg.textContent= 'Number of parts= ' + tmp + '. If raster is higher, the stimulus will be broken down into a larger number of parts and vice versa';
        //msg.textContent+= ' Lower value = task is easier. Higher value = the task is more difficult';    
			}
			else  {
				msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 					
			break; 

    case 'rotation':   
			if (e.checked)	{
				msg.textContent= 'Parts of the stimulus will be rotated';   
				document.getElementById('is_border').disabled= true;
			}
			else	{
				msg.textContent= 'The stimulus parts are not rotated'; 
				document.getElementById('is_border').disabled= false;
			}
			
			break; 

		case 'mirror':   
			msg.textContent= e.checked ? 'The left and right parts of the image around the center axis within each row are replaced' : 'Stimulus segments are randomly scattered';    
			break;
		
    case 'demo':
    case 'is_border':		
			msg.textContent= e.checked ? 'Yes' : 'No';    
			break;
    		
		case 'feedback':   
			msg.textContent= e.checked ? 'Additional information will be displayed at the bottom of the page.' : 'No additional information is displayed at the bottom of the page.';    
			break;
						 
		case 'back_clr': 
			if (validColor(e.value))   {           
				msg.removeAttribute('class');       
				msg.textContent= (document.getElementById('stype').value === "1") ? 'Background color for image stimuli. Adjust it to your image stimuli.' : 'Disabled for alphanumeric characters (default = white). Enabled for image stimuli only.';  
			}
			else  {
				msg.textContent= 'The background color is invalid.';
				msg.setAttribute('class', 'error');    // msg.className = 'error';
			} 
			break; 
		 
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// check if items count is valid or not
function itemsValid()  
{
  let e= document.getElementById('item_cnt'), msg= document.getElementById('item_cnt_msg'), s = 'Values within this range are allowed: ', s2= ' with step= '; 
	let stim_type= document.getElementById('stype').value;  // stimulus type
	let images= document.getElementById('preview').getElementsByTagName('img');  
	let img_cnt = (images && images.length) ? images.length : 0;
	if ( ( e.validity.valid === false ) ||
       (( stim_type === '1' && parseInt(e.value, 10) > img_cnt ) ) )  {
		msg.textContent= s + e.min + '..' + e.max + s2 + e.step;
		if ( stim_type === '1') {
			msg.textContent+= '. For Stimulus type Images: Item count can not be greater then images count. Images count=' + img_cnt; 
    }		
		msg.setAttribute('class', 'error');    // msg.className = 'error';
		return;		
	}	
	msg.removeAttribute("class");       
	msg.textContent= (stim_type === '1') ? 'The number of stimuli (images) in the task' : 'The number of stimuli (random alphanumeric characters) in the task';   
}  


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  (dirty===true) ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// check for validity.valid object in all input type="number" + other internal app checks.see: setPar()
function optionsValid()  
{     
  let res= document.getElementById('item_cnt').validity.valid && document.getElementById('s_parts').validity.valid;
	if (document.getElementById('stype').value == '0' && document.getElementById('alfa_cnt').validity.valid == false)	{ 
		res= false; 
	}	
	if (document.getElementById('stype').value == "1" && document.getElementById('preview').getElementsByTagName('img').length === 0) {	
	  res= false;
	}
  return res;               
}


// start task (research) 
function startTask(e)   
{     
  if (stype === '1' && document.getElementById('preview').getElementsByTagName('img').length === 0) {		 
		document.getElementById('is_valid').textContent= 'Unable to start task because you did not select image files to use as stimuli. Number of images = 0. Go to "Options" => "Select image files".'; 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold');
    return;		
	}  
	
	setPar();
 
	setStim();
	
	document.getElementById('result_wrapper').style.display="none";
  document.getElementById('end').style.display= "none"; 
  document.getElementById('start').style.display= "none";              
  document.getElementById('footer').style.display= "block"  	
	document.getElementById('progress').style.display= "block"; 
  document.getElementById('progress').setAttribute("style", "width:100%");
	document.getElementById('is_correct').textContent= '';
	if (feedback)  {
		document.getElementById('is_correct').style.display= 'block';
		document.getElementById('items_left').style.display= 'block';
	}
	else  {
		document.getElementById('is_correct').style.display= 'none';
		document.getElementById('items_left').style.display= 'none';
	}
	canvas.style.display= 'block'; 
	canvas.style.backgroundColor= '#ffffff';
  document.getElementById('back_canvas').style.backgroundColor= '#ffffff';
	ctx.strokeStyle= '#000000';
	if (stype === '1' && validColor(back_clr) ) 	{
		try  {
			canvas.style.backgroundColor= back_clr;
      document.getElementById('back_canvas').style.backgroundColor= back_clr;
		}
		catch(e) {			
		}				
	}	  
	if (stype === "0")
	{
		switch (alfa_type)
		{		
			case '1':  // Letters-lowercase 
				document.getElementById('answer').setAttribute('style','text-transform:lowercase');			
				break;
			case '2':  // Letters-UPPERCASE 
				document.getElementById('answer').setAttribute('style','text-transform:uppercase');			
				break;
			default:  // Digits ("0") or All digits and letters ("3")
				document.getElementById('answer').removeAttribute('style');	
		} 
	}
	else  {
		document.getElementById('answer').removeAttribute('style');	
		//document.getElementById('answer').setAttribute('style','text-transform:lowercase');		
	}
	
	is_task= true;
	
	setPage();      
	
	drawStim(true);     // show (draw) first stimulus
      
}


// set internal application research parameters
function setPar()  
{  
  stim= [];   
	ans= [];  //  user answers for each item (dependent var)           
  dv= [];
  
  for (let j= 0; j< item_cnt; j++)   {
    ans[j]= '';
    dv[j]= 0;  // The stimulus is not recognized even if it is displayed in its original form (consisting of only 1 part)
  }
	item= 0;
  c_parts= s_parts;  // current level == start level
	if (stype === '1' && item_cnt > document.getElementById('preview').getElementsByTagName('img').length) {		 
		item_cnt= document.getElementById('preview').getElementsByTagName('img').length;
		document.getElementById("is_valid").textContent= "The number of items was greater than the number of images. The number of items is automatically reduced to: " + item_cnt; 
    document.getElementById("is_valid").setAttribute("style", "text-align:center;color:#ff0000;font-weight:bold"); 
	}  
}


// generates task stimuli 
function setStim()	
{
	if (stype === '0')  // alphanumeric characters are selected
	{  
		let s, previous;		
		for (let j= 0; j<item_cnt; j++)   
		{		
			do
			{	
			  s='';
				switch (alfa_type)
				{
					case '0':  // digits 		
						for (let i=0; i<alfa_cnt; i++)  {
							s+= getRndInt(0,9).toString();
						}			  
						break;			
					case '1':  // Letters-lowercase 
						s= lowerStrGen(alfa_cnt);			
						break;
					case '2':  // Letters-UPPERCASE 
						s= upperStrGen(alfa_cnt);			
						break;
					default:  // All digits and letters 
						s= allStrGen(alfa_cnt);
				} 
			} while (s === previous);
			stim[j]=s;
			previous=s;		
		}
	}
	else  // images are selected
	{
		let image;
		for (let j= 0; j<item_cnt; j++)   {
			image= document.getElementById('preview').getElementsByTagName('img')[j];
			stim[j]= image.title ? image.title : 'image ' + parseInt(j+1);  // images file names array => document.querySelector('input[type=file]').files[i].name
			//let is_title= image.getAttribute('data-title');
			//stim[j]= is_title ? is_title : 'item ' + parseInt(j+1);  // images file names array => document.querySelector('input[type=file]').files[i].name
		}					
	}
}


// generates a random string from lowercase letters
function lowerStrGen(cnt)
{ 
  let all = "abcdefghijklmnopqrstuvwxyz";
  let txt = "", tmp= all.length;
  for (let i = 0; i < cnt; i++)  {
    txt += all.charAt(Math.floor(Math.random() * tmp));
	}
  return txt;
}


// generates a random string from uppercase letters 
function upperStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  let txt = "", tmp= all.length;
  for (let i = 0; i < cnt; i++)  {
    txt += all.charAt(Math.floor(Math.random() * tmp));
	}
  return txt;
}


// generates a random string from all (uppercase and lowercase) letters and digits
function allStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
  let txt = "", tmp= all.length;
  for (let i = 0; i < cnt; i++)  {
    txt += all.charAt(Math.floor(Math.random() * tmp));
	}
  return txt;
}


// set canvas.height, canvas.width, global font_size
// call in startTask() + each body onresize() when is_task == true => call this function
function setPage()  
{			
	// The read-only innerHeight property of the Window interface returns the interior height of the window in pixels, including the height of the horizontal scroll bar, if present.
  // window.innerHeight = The value of innerHeight is taken from the height of the window's layout viewport.  
	
	let h= window.innerHeight - document.getElementById('footer').offsetHeight; // window height - footer height
  canvas.style.height= h + 'px'; // The height property has effect only on block-level elements or on elements with absolute or fixed position.	
  canvas.height= h;
	canvas.width= window.innerWidth; 
	
	// set global canvas font size (px) for drawText() function
  font_size= Math.round(Math.min(Math.floor(canvas.width/alfa_cnt), 0.5 * canvas.height));		
	
	drawStim(false);
}	


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
   
	if (document.getElementById('item_cnt').validity.valid)  {  
    if ( stype === '0' || (stype === '1' && parseInt(document.getElementById('item_cnt').value, 10) <= document.getElementById('preview').getElementsByTagName('img').length) ) {
			item_cnt= parseInt(document.getElementById('item_cnt').value, 10);     
		}
  } 
	 	 
  stype= document.getElementById('stype').value;     
	
  alfa_type= document.getElementById('alfa_type').value;
	
	if (document.getElementById('alfa_cnt').validity.valid) {
		alfa_cnt= parseInt(document.getElementById('alfa_cnt').value, 10);
  }		
	
	font_family= document.getElementById('font_family').value;
		    
  if (document.getElementById('s_parts').validity.valid)  {   
    s_parts= parseInt(document.getElementById('s_parts').value, 10);   
	}
	
	demo= document.getElementById('demo').checked ? true: false; 
	
	feedback= document.getElementById('feedback').checked ? true: false; 
	
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
		     
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  
}


// get random integer from min to max range
function getRndInt(min, max) {
	return Math.floor(Math.random() * (max - min + 1)) + min;
}


// get random float from min to max range
function getRndFloat(min, max) {
  return Math.random() * (max - min) + min;
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// go to the next task if the ENTER key is pressed. see function init()
function answerKeyUp(e)  
{
	if (e.key === 'Enter' || e.keyCode === 13) {  
		next();
	}
}


// goto next item (stimulus)
function next()  
{  
  ans[item]= document.getElementById('answer').value;	
  // The \s meta character in JavaScript regular expressions matches any whitespace character: spaces, tabs, newlines and Unicode spaces. And the g flag tells JavaScript to replace it multiple times. 
	ans[item]= ans[item].replace(/\s/g, '');	 // const newStr = str.replace(regexp|substr, newSubstr|function)
	if (stype === '0')	{
		if (alfa_type === '1')  {
			ans[item]= ans[item].toLowerCase(); // "1" = Letters-lowercase;
		}	
		else if (alfa_type === '2')  {
			ans[item]= ans[item].toUpperCase(); //  "2" =Letters-UPPERCASE;
		}	  // alfa_type => "3" is case sensitive (all letters and digits); "0" => Digits
	}
	// Final evaluation of the entered answer
	let is_cor= false; // answer is not correct 
  if ( (stype === '0' && ans[item] == stim[item]) || (stype === '1' && ans[item].toLowerCase() == 'y') )	{
		is_cor= true; // answer is correct
    dv[item]= c_parts * c_parts; // For each stimulus: the number of parts (raster) on which the user recognized the stimulus.    
	}
			
	if (is_cor || c_parts <= 1)	{  // goto next stimulus
		c_parts= s_parts;
		item++;
	}	
	else  {
		c_parts--;
	}	
	if (feedback)  {
	  document.getElementById('is_correct').textContent= (c_parts == s_parts) ? 'A new stimulus is displayed.' : 'The same stimulus is displayed with a smaller number of parts.'; 
		document.getElementById('is_correct').textContent+= is_cor ? ' The last answer was correct.' : ' The last answer was incorrect.';
	}	 
	if (item < item_cnt)  {         		 
    drawStim(true);     // draw next stimulus => range 0..item_cnt-1        
  }
  else  {               
    theEnd();  // end of task if item >= item_cnt 
  }
}


/* draw (repaint) stimulus
   upd_ans == true => update current answer.  upd_ans == false => calls from body onresize="pageResize()" => setPage() */     
function drawStim(upd_ans) 
{  
  	
	clear();  // clear canvas
	
	if (stype === '0')
	{     		
		drawText(stim[item]); 		
		setTextBounds();
    canvasToImage();				
	}
	else	
	{		
    document.getElementById('img_cont').src= document.getElementById('preview').getElementsByTagName('img')[item].src;    		
		//document.getElementById('img_cont').crossOrigin= 'anonymous'; // If crossOrigin is an empty string (""), the anonymous mode is selected.		 		
	}		
 	
	if (upd_ans)  {  
		document.getElementById('answer').value=  (demo && stype === '0') ?  stim[item] : '';  // show correct answer or delete previous answer
		document.getElementById('answer').focus();
		updateProgress();
	}
		  
}


// draw text (stimulus) on canvas 
function drawText(txt)  
{
	//ctx.fillStyle = "#ffffff"; 
	//ctx.clearRect(0, 0, canvas.width, canvas.height);      // clear canvas
	let x= canvas.width/2;
	let y= canvas.height/2;	     
	ctx.fillStyle = "#000000";  // line color 	
	let dx= Math.floor(0.5 * canvas.width);  
	let dy= Math.floor(0.5 * canvas.height); 
	let rad= Math.floor(Math.min(dx,dy)/2);  // text radius 	                                                                   
  /* https://developer.mozilla.org/en-US/docs/Web/CSS/font-family   
	   https://www.w3schools.com/cssref/css_websafe_fonts.asp	 
		 examples: ctx.font = "72px monospace, Courier";  	ctx.font = "5em Georgia";  
	*/
	ctx.font= font_size.toString() + 'px ' + font_family;  
	ctx.textBaseline= 'bottom';  
  ctx.textBaseline= 'middle';    
	ctx.textAlign= 'center';
	
	ctx.fillText(txt, x, y); 	
}	


/* draw image from ../images directory into canvas object (html container)
   see:  https://coderwall.com/p/jzdmdq/loading-image-from-local-file-into-javascript-and-getting-pixel-data
	 https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/drawImage
	 https://stackoverflow.com/questions/17049176/ns-error-not-available-component-is-not-available?rq=1  */
function loadImage()  
{	
	let max_w= Math.round(0.95*canvas.width), max_h= Math.round(0.95*canvas.height);
	let img= document.getElementById('img_cont');  // image stimulus html container
	let x0, y0, w, h;
	w= img.naturalWidth;
	h= img.naturalHeight;		
	if (img.naturalWidth <= max_w && img.naturalHeight <= max_h)  
	{ 
		x0= Math.floor(canvas.width/2 - img.naturalWidth/2);
		y0= Math.floor(canvas.height/2 - img.naturalHeight/2);					
	}
	else  
	{  	// rescale image			
		while (w > max_w || h > max_h)  {
			w= w/2;  // decrease width 100%
			h= h/2;  // decrease height 100%
		}  				
		x0= Math.floor(canvas.width/2 - w/2);
		y0= Math.floor(canvas.height/2 - h/2);						
	}	
	img.width= w;
	img.height=h;
	//img.style.width= w + 'px';  
	//img.style.height= h + 'px';	 			
	bounds ={left:x0, top: y0, width: w, height: h};	// canvas bounds

	if (!is_task) {
	  return;	
	}	
  
  clear();  // clear canvas
		
	if (c_parts <= 1)  { 
		ctx.drawImage(img, x0, y0, w, h);	
	}
	else  
  {		 			
    if (document.getElementById('rotation').checked)  {
      puzzleRotate(document.getElementById('mirror').checked);  //puzzleRotate(true||false);
    }
    else  {
      puzzleImage(document.getElementById('mirror').checked);   //puzzleImage(true||false);
    } 				
	}		
}


// Update Progress bar on footer
function updateProgress()
{
  let percent = Math.round (100* (item_cnt - item) / item_cnt);       
	document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
	document.getElementById('items_left').textContent= feedback ? 'Item: ' + (item + 1) + '. Number of parts: ' + (c_parts * c_parts) + '. Remaining items: ' + (item_cnt - item-1) : '';
  if (item === item_cnt-1)	{		
		document.getElementById('next').style.backgroundImage = "url('img/end.png')";
		document.getElementById('next').title = 'End of task';
	}
	else	{
		document.getElementById('next').style.backgroundImage = "url('img/forward.png')"; 
		document.getElementById('next').title = 'Next task';
	}
}	


// end of task (research)
function theEnd()  
{   
  is_task= false;  
 	clear();
	canvas.style.display= 'none';
	document.getElementById('footer').style.display= 'none';
  document.getElementById('end').style.display= 'block';   
  
	// compute total number of recognized stimulus (correct) 
  let cor = 0, rows = [], dv_ok= [];   
	for (let i=0; i<item_cnt; i++)  
	{
		if ( (stype === '0' && ans[i] == stim[i]) || (stype === '1' && ans[i].toLowerCase() == 'y') )  {
      cor++;  // number of recognized stimulus (correct answers)
      dv_ok.push(dv[i]);
      rows[i]= [ parseInt(i+1,10), stim[i], ans[i], 'Yes', dv[i] ];  // => ['Item', 'Stimulus' , 'Answer', 'Correct=Yes', 'Number of parts']); 	        
		}
		else  {  // unrecognized stimulus
      rows[i]= [ parseInt(i+1,10), stim[i], ans[i], 'No', '' ];  // => [ 'Item', 'Stimulus' , 'Answer', 'Correct=No', ""]); 	        
		}	   
	}		
				
  document.getElementById('result_wrapper').style.display= 'block';  
	createTableArray(document.getElementById('table1'), rows,  ['Item', 'Stimulus' , 'Answer', 'Correct', 'Number of parts' ]);  
	document.getElementById('table1').style.display= 'block';  	
	
	// calculate decriptive statistics for the number of parts (raster) on which the user recognized the stimulus => average and sample standard deviation..	

  let m = (cor > 0) ? mean(dv_ok).toFixed(3) :  'Cannot be calculated';  // mean for number of parts (raster) on which the user recognized the stimulus.
	let sd = (cor > 1) ? stDev(dv_ok, m).toFixed(3) : 'Cannot be calculated'; // standard deviation for number of parts (raster) on which the user recognized the stimulus. 
  
	let tmp=[];
	tmp[0]= ['Number of parts', m, sd, cor];		
	createTableArray(document.getElementById('table2'), tmp, ['Variable', 'Average', 'Sample Standard Deviation', 'N']);   
		
	try 
  {
    if (store) {
			if ( isNaN(m) ) { m= '';  }   // recode 'Cannot be calculated' => '' for browser storage
      if ( isNaN(sd) ) { sd= '';  } // recode 'Cannot be calculated' => '' for browser storage
			setStorage(cor, m, sd);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }    	

}


// return arithmetic mean (mean, average) of array (arr) 
function mean(arr)  
{	
	let sum=0;
	for (let i in arr)  {
		sum+= parseFloat(arr[i]);
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation
// arr = array argument; mean = average (mean)
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
	  sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* set user data into window.localStorage
	 cor  => number of recognized stimulus (correct answers)
	 m => arithmetic mean for the number of parts (raster) on which the user recognized the stimulus.
	 sd => sample Standard Deviation for the number of parts (raster) on which the user recognized the stimulus. */ 
function setStorage(cor, m, sd)   
{  
  let oldItems = JSON.parse(localStorage.getItem("wm14-v1.0")) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id': id,
      'date': setDate(),   //new Date(),  
			'items': item_cnt,     
      'correct': cor,                   
      'Mean': m, 
			'SD': sd,  			      
			'stimulus-type' : stype,
			'start-level': s_parts,
			'alfa-type': stype === '0' ? alfa_type : '', 
			'alfa-cnt':  stype === '0' ? alfa_cnt  : '', 
			'rotation': document.getElementById('rotation').checked ? '1' : '0',
			'mirror': document.getElementById('mirror').checked ? '1' : '0',			
			'feedback': feedback ? '1' : '0',
      'demo' : (demo && stype != "1") ? "1": "0"        
  };

  oldItems.push(newItem);

  localStorage.setItem("wm14-v1.0", JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems); 
  	
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear wm14 web browser storage data
function clrStorage()  
{  
  let cnt= localStorage.getItem('wm14-v1.0')  ?  JSON.parse(localStorage.getItem('wm14-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no wm14 data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' wm14 results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved wm14 results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('wm14-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data  ';
 }  
}


// clear the canvas
function clear()  {	 
	//ctx.fillStyle = "#ffffff";
	ctx.clearRect(0, 0, canvas.width, canvas.height);
	//ctx.fill();   
}	


/* EventListener for input type="file" onchange. see html: document.getElementById("add_file")
   When the user selects image files
   https://developer.mozilla.org/en-US/docs/Web/API/FileReader/readAsDataURL */
function createImages() 
{
  let preview = document.querySelector('#preview');
  let files = document.querySelector('input[type=file]').files;
  let cnt= document.getElementById('preview').getElementsByTagName('img').length;  
	function readAndPreview(file) 
	{    
    if ( /\.(jpe?g|png|gif|bmp)$/i.test(file.name) ) {  // Make sure `file.name` matches our extensions criteria
      let reader = new FileReader();
      reader.addEventListener("load", function () {
        let image = new Image();              
        image.src = this.result;        
				image.setAttribute('class', 'flex-child');   //image.height = 100;
				image.title = 'item ' + parseInt(cnt+1) + '=' + file.name;
				//image.setAttribute('data-title', 'item ' + parseInt(cnt+1)); 
				// drag and drop  
				image.id= 'img' + parseInt(cnt+1); 
				image.setAttribute("draggable", "true"); 
				image.setAttribute("ondragstart", "onDragStart(event)");   				
				image.setAttribute("ondragover", "onDragOver(event)");
				image.setAttribute("ondrop", "onDrop(event)");
				
				preview.appendChild(image);
				cnt++;
				if (cnt>0) document.getElementById('img_info').style.display= 'block';
				document.getElementById('item_cnt').value= cnt; 
				itemsValid();   //document.getElementById('item_cnt_msg').textContent= '';
        refMsg();
      }, false);
      reader.readAsDataURL(file); 
    }
  }
  if (files) {
    [].forEach.call(files, readAndPreview);
  }
  dirty=true;  
	//document.getElementById('item_cnt').value= files.length;       	
}


// ondragstart event for stimulus type= images
function onDragStart(ev)  {   
  ev.dataTransfer.setData('text', ev.target.id);
}


// ondragover() event for stimulus type= images
function onDragOver(ev)  {  
  ev.preventDefault();
}


// ondrop() event for stimulus type= images
function onDrop(ev)  
{  
	ev.preventDefault(); 
  let my_id = ev.dataTransfer.getData('text');    // node1.id
  let node1= document.getElementById(my_id);    // node1= DRAG node pointer           
	let node2= ev.target;   //  node2= DROP node pointer: above  <ol> or <li> ??  => [object HTMLOListElement] or [object HTMLLIElement] => this is drop node      
	let parNode= document.getElementById('preview');  
  if ( node2.tagName === "IMG" )  {
    if ( node2 === parNode.lastChild ) {
      insertAfter(node1, node2);  // append image to end of list
    } 
    else {
      node2.parentNode.insertBefore(node1, node2);   // move (insert) node1 before node2 => let insertedNode = parentNode.insertBefore(newNode, referenceNode)
    }	    
  }
  toogleInfo();
}


// refNode is the node you want to put newNode after.
function insertAfter(newNode, refNode) {
  refNode.parentNode.insertBefore(newNode, refNode.nextSibling);
}


// remove image file from container (stimulus type: images). see html source
function delItem(ev)
{ 
	ev.preventDefault(); 
  let my_id= ev.dataTransfer.getData('text');    // node1.id
  let node1= document.getElementById(my_id);    // node1= DRAG node pointer             		
	node1.parentNode.removeChild(node1);   // delete ex node2
	toogleInfo();
  dirty=true;  
  refMsg();	
}


// auto set item_cnt in task based on current number of images stimulus
// set visibility for small info div: "Drop image in this line here to delete" and "Use drag and drop to move files (stimulus)"
function toogleInfo()  
{
	let img_cnt= document.getElementById('preview').getElementsByTagName('img').length;  // current number of images stimulus
	document.getElementById('item_cnt').value= img_cnt;
  document.getElementById('img_info').style.display= img_cnt > 0 ? 'block' : 'none';
	itemsValid();
}	


// creates a set of rectangles within an canvas bounds. raster => c_parts * c_parts
function setRects()
{
	rects = [];  	
	let dx = bounds.width / c_parts, dy = bounds.height / c_parts, i = 0, off;
	if (stype === '1' && document.getElementById('is_border').checked && !document.getElementById('rotation').checked)  {
		off= 1;    //  only for image with no rotation => off == 1 => add 1px offset (border) between rects (image parts)  
	}
	else  {
		off= 0;  //  off == 0 => no offset (border) between rects (image parts)
	}
	for (let row = 0; row < c_parts; row++ )
	{	    	
		for (let col = 0; col < c_parts; col++)
		{													
			let x = bounds.left + col * (dx + off);
			let y = bounds.top  + row * (dy + off);  
			let	w = dx;
			let	h = dy;
			rects.push({
				x0:x,         
				y0:y, 
				w: dx,
				h: dy,
				full: 0   
			});			     				
		}		
	}	
}


// puzzle image with raster of rectangles (no rotation). raster => c_parts * c_parts
function puzzleImage(is_mirror)
{
	setRects(); 	 
	let img= document.getElementById('img_cont');  // image stimulus html container 	
  let dx= img.naturalWidth / c_parts;   
	let dy= img.naturalHeight / c_parts;  
	let  i= 0, sel = {};   //  i => rects (image parts) counter;  sel => previously selected image rects 				
	if (is_mirror) 
	{  // puzzle mirror rectangles replace within each row: rect left <=> right rect
		for (let row = 0; row < c_parts; row++ )
		{	    	
			for (let col = c_parts-1; col > -1; col--)
			{							
				ctx.drawImage(img, col*dx, row*dy, dx, dy,   rects[i].x0, rects[i].y0, rects[i].w, rects[i].h);	
				i++;  //void ctx.drawImage(image, sx, sy, sWidth, sHeight, dx, dy, dWidth, dHeight);		
			}
		}					
	}
  else
	{  // random puzzle generator
		let last= rects.length-1, j, not_ok;
		for (let row = 0; row < c_parts; row++ )
		{	    	
			for (let col = 0; col < c_parts; col++)
			{
				do  {
					j = getRndInt(0, last);		// choose image random rect region			
					not_ok = (typeof sel[j] != "undefined") || rects[j].full != 0 || ( j == i && i < last);			
				}	while ( not_ok );
				sel[j]= 1; 
				rects[j].full= 1;							  	
				ctx.drawImage(img, col*dx, row*dy, dx, dy,   rects[j].x0, rects[j].y0, rects[j].w, rects[j].h);	
        i++;  //void ctx.drawImage(image, sx, sy, sWidth, sHeight, dx, dy, dWidth, dHeight);					
			}		
		}   	
	}	
}	


/* puzzle image with raster of rectangles. rotate each rectangle within image.
   is_mirror == true => puzzle mirror generator is used => within each row each rectangle left is changed with mirror right rectangle
	 is_mirror == false => puzzle pseudorandom generator is used  */
function puzzleRotate(is_mirror)
{ 
	setRects();
	let img= document.getElementById('img_cont');  // image stimulus html container 
	let dx= img.naturalWidth / c_parts;    
	let dy= img.naturalHeight / c_parts;   		
	let off = 1, i = 0, sel = {}, scale;   //  off == 0 => no offset between rects (image parts); sel => previously selected image rects  
	if (document.getElementById('rotation').checked)  {				
		let d= Math.sqrt ( Math.pow(dx, 2) + Math.pow(dy, 2) );    // d = hypotenuse = the length of the diagonal by the Pythagorean theorem.  Math.pow(base, exponent)		
		scale= Math.min(dx, dy) / d;  // axis scale factor =>  min(width,height)/ hypotenuze length
	}
	else {
		scale=1; 
	}	  
	if (is_mirror)  
	{  // puzzle mirror rectangles replace within each row: rect left <=> right rect
		for (let row = 0; row < c_parts; row++ )
		{	    	
			for (let col = c_parts-1; col > -1; col--)
			{		        
				let x= rects[i].x0+ rects[i].w/2, y= rects[i].y0 + rects[i].h/2;				
				ctx.save();  //ctx.translate(x,y);  				
				ctx.setTransform(scale, 0, 0, scale, x, y);
				let alpha= getRndFloat(0, 2*Math.PI); 
				ctx.rotate(alpha);  //ctx.rotate(Math.PI/4);
				ctx.drawImage(img, col*dx, row*dy, dx, dy,   rects[i].x0-x, rects[i].y0-y, rects[i].w, rects[i].h);  //void ctx.drawImage(image, sx, sy, sWidth, sHeight, dx, dy, dWidth, dHeight);	
				ctx.restore();							
				i++;  	
			}			
		}					
	}
  else
	{  // pseudorandom puzzle generator
		let last = rects.length-1, j, not_ok;
		for (let row = 0; row < c_parts; row++ )
		{	    	
			for (let col = 0; col < c_parts; col++)
			{
				do  {
					j = getRndInt(0, last);		// choose image random rect region			
					not_ok = (typeof sel[j] != 'undefined') || rects[j].full != 0 || ( j == i && i < last);							
				}	while (not_ok);
				sel[j]= 1; 
				rects[j].full= 1;	        
				let x= rects[j].x0+ rects[j].w/2, y= rects[j].y0 + rects[j].h/2;				
				ctx.save();  //ctx.translate(x,y); 
				ctx.setTransform(scale, 0, 0, scale, x, y);
				let alpha= getRndFloat(0, 2*Math.PI);   
				ctx.rotate(alpha);  //ctx.rotate(Math.PI/4);
				ctx.drawImage(img, col*dx, row*dy, dx, dy,   rects[j].x0-x, rects[j].y0-y, rects[j].w, rects[j].h);
				ctx.restore();							        
				i++;				
			}		
		}   	
	}
  ctx.setTransform(1, 0, 0, 1, 0, 0); // Reset transformation matrix to the identity matrix
}	


// set canvas bounds for output of drawText()
function setTextBounds()  
{
  
  let imageData= ctx.getImageData(0, 0, canvas.width, canvas.height);		

	let res= getBlackRect();  // res => ( [0] = first row = y, [1] = first column = x, [2] = last row = y, [3]= last column= x) with black pixels   	

	let off= 10; // offset around text (px)
	
  bounds= {left:res[1]-off, top: res[0]-off, width:res[3]-res[1]+2*off, height: res[2]-res[0]+2*off};	//bounds ={left:x0, top: y0, width: w, height: h};		
  
}


/* use only for black pixels search on window.canvas object (html canvas element)
   return rect bounds ( [0] = first row = y, [1] = first column = x, [2] = last row = y, [3]= last column= x) with black pixels 
   return [left, top, right, bottom] corners for black pixels */
function getBlackRect()
{
  let pixels= ctx.getImageData(0, 0, canvas.width, canvas.height).data;
  let cnt= 0, row= 0, col= 0; 
	let res=[-1, -1, -1, -1];  // left, top, right, bottom corners
  let len= pixels.length;
	let w= canvas.width;	
	for (let i = 0; i != len; i += 4) 
	{		
		if ( res[0] == -1 && pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[0]= row;   // find first row with minimum one black pixel   
		}
    else if ( pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[2]= row;   // update (this) last row with minimum one black pixel   
		}			
		if ( (res[1] == -1 || col < res[1]) && pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[1]= col;  // find first column with minimum one black pixel  
		}
		else if ( col > res[3] && pixels[i] == 0 && pixels[i+1] == 0 && pixels[i+2] == 0 && pixels[i+3] > 0 ) {
			res[3]= col;  // find last column with minimum one black pixel  
		}	
	  cnt++;    // total pixel counter
		col++;    // column counter
		if ( cnt % w == 0 ) {   // enter new row
			row++;	
			col=0;
		}	
	}
	row--;			
	return res;
}	


// copy a part of canvas to image
function canvasToImage()
{  
  let imgData = ctx.getImageData(bounds.left, bounds.top, bounds.width, bounds.height);  // get alfanumeric text content
  clear();  // make invisible alfanumeric stimulus early as possible
	document.getElementById('img_cont').src = getImageURL(imgData, bounds.width, bounds.height);      
}	


// copy a part of canvas to image
// https://stackoverflow.com/questions/45234492/copy-a-part-of-canvas-to-image
function getImageURL(imgData, width, height) {
   let backCanvas= document.getElementById('back_canvas')
   let ctx = backCanvas.getContext('2d');
   backCanvas.width = width;
   backCanvas.height = height;    //ctx.fillStyle = "#ffffff";
	 ctx.clearRect(0, 0, width, height);   
   ctx.putImageData(imgData, 0, 0);
   return backCanvas.toDataURL();  //image URL
}
